import inspect
import json
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar, Union
from warnings import warn

try:  # pragma: no cover
    from typing import get_args, get_origin
except ImportError:  # pragma: no cover
    from typing_inspect import get_args, get_origin  # type: ignore

from .errors import SchemaWarning
from .field import Field


def _type_name(t):
    if isinstance(t, type):
        return t.__name__
    else:
        return type(t).__name__


BUILT_IN_TYPES = {
    object: (
        "object",
        "An arbitrary, flexible type determined at runtime (can be str, int, float, bool, list, object)",
    ),
    str: ("String", "A UTF-8 encoded string"),
    int: ("Integer", "An integer"),
    float: ("Float", "A floating point number"),
    bool: ("Boolean", "A boolean, true or false"),
}


def _get_type_info(t: type, types_by_class: dict, types_by_name: dict) -> dict:
    t, is_optional = check_optional(t)

    if is_optional:
        return {"kind": "NON_NULL", "ofType": _get_type_info(t, types_by_class, types_by_name)}
    elif check_list(t):
        item_type = get_list_type(t) or object
        return {"kind": "LIST", "ofType": _get_type_info(item_type, types_by_class, types_by_name)}
    else:
        introspected = introspect_type(t, types_by_class, types_by_name)
        return {"kind": introspected["kind"], "name": introspected["name"]}


def _parse_arg_docs(doc: Optional[str]) -> dict:
    if not doc:
        return {}

    arg_docs = {}
    found_args = False

    for line in doc.splitlines():
        if line.strip() == "Args:":
            found_args = True
        elif found_args and ":" in line:
            arg_name, arg_doc = line.strip().split(": ", 1)
            arg_docs[arg_name] = arg_doc

    return arg_docs


def _parse_default_args(resolver: Callable) -> Dict[str, Any]:
    arg_spec = inspect.getfullargspec(resolver)
    arg_defaults = arg_spec.kwonlydefaults or {}

    positional_args = reversed(arg_spec.args)
    positional_defaults = reversed(arg_spec.defaults or [])
    arg_defaults.update(zip(positional_args, positional_defaults))

    return arg_defaults


def check_optional(t: type) -> Tuple[type, bool]:
    args = get_args(t)
    if get_origin(t) is Union and len(args) == 2 and isinstance(None, args[1]):
        return args[0], True
    else:
        return t, False


def check_list(t: type) -> bool:
    return t is list or get_origin(t) is list


def get_list_type(t: type) -> Optional[type]:
    args = get_args(t)
    if args:
        return args[0]
    else:
        return None


def _get_args(field: Field, types_by_class: dict, types_by_name: dict) -> List[dict]:
    if not field.args:
        return []

    arg_docs = _parse_arg_docs(field.doc)
    default_values = _parse_default_args(field.resolver)

    introspected_args = []
    for arg, arg_type in field.args.items():
        arg_type, is_optional = check_optional(arg_type)
        type_info = _get_type_info(arg_type, types_by_class, types_by_name)

        if arg in default_values:
            default: Optional[str] = json.dumps(default_values.get(arg))
        else:
            default = None

        if not is_optional and default != "null":
            type_info = {"kind": "NON_NULL", "ofType": type_info}

        introspected_args.append(
            {
                "name": arg,
                "description": arg_docs.get(arg),
                "type": type_info,
                "defaultValue": default,
            }
        )

    return introspected_args


def _get_field_description(doc: Optional[str]) -> Optional[str]:
    if not doc:
        return doc

    description = []
    for line in doc.splitlines():
        line = line.strip()
        if line == "Args:":
            break
        elif line:
            description.append(line)

    return "\n".join(description)


def introspect_type(t, types_by_class: dict, types_by_name: dict, is_operation: bool = False) -> dict:
    if isinstance(t, TypeVar):  # pragma: no cover
        t = object

    try:
        return types_by_class[t]
    except KeyError:
        pass

    try:
        name, doc = BUILT_IN_TYPES[t]
        built_in_info = {"kind": "SCALAR", "name": name, "description": doc}

        types_by_class[t] = built_in_info
        types_by_name[name] = built_in_info
        return built_in_info
    except KeyError:
        pass

    fields = []
    for attr in dir(t):
        field = getattr(t, attr)
        if not isinstance(field, Field):
            continue

        type_info = _get_type_info(field.field_type, types_by_class, types_by_name)
        args = _get_args(field, types_by_class, types_by_name)
        description = _get_field_description(field.doc)

        fields.append({"name": field.name, "description": description, "args": args, "type": type_info})

    if is_operation and not fields:
        return None  # type: ignore

    extra = {}
    if is_operation or fields:
        kind = "OBJECT"
    elif inspect.isclass(t) and issubclass(t, Enum):
        kind = "ENUM"
        extra["enumValues"] = [{"name": e.name} for e in t]
    else:
        kind = "SCALAR"

    name = _type_name(t)
    introspected = {"kind": kind, "name": name, "fields": fields, "description": t.__doc__, **extra}

    if kind == "OBJECT":
        introspected["interfaces"] = []

    types_by_class[t] = introspected

    if name in types_by_name:
        warn(f"Multiple types found with the same name during introspection: {name}", SchemaWarning)
    else:
        types_by_name[name] = introspected

    return introspected



from ast import literal_eval
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from parsimonious import Grammar, NodeVisitor  # type: ignore

# Note: null and enum have "" suffix so that the visitor function is
# called, see https://github.com/erikrose/parsimonious/issues/111
grammar = Grammar(
    r'''
document                 = _? (operation _?)+

operation                = named_explicit_operation / explicit_operation / fields
explicit_operation       = operation_type operation_variables? _? fields
named_explicit_operation = operation_type _ identifier operation_variables? _? fields
operation_variables      = _? "(" _? operation_variable (_ operation_variable)* _? ")"
operation_variable       = "$" identifier _? ":" _? type (_? "=" _? value)?
operation_type           = "query" / "mutation"

fields                   = "{" _? field (_ field)* _? "}"
field                    = field_alias? identifier field_arguments? (_? fields)?
field_alias              = identifier _? ":" _?
field_arguments          = _? "(" _? pair (_ pair)* _? ")"
pair                     = identifier _? ":" _? value
value                    = list / object / boolean / null / string / number / integer / variable / enum

boolean                  = "true" / "false"
null                     = "null" ""
string                   = ~r'"""(\\.|(?!""").)*"""'s / ~r'"(\\[^\n]|[^\\"\n])*"'
variable                 = "$" identifier
enum                     = identifier ""
integer                  = ~"-?(0|[1-9][0-9]*)"
number                   = scientific / decimal
decimal                  = ~r"-?[0-9]+\.[0-9]+"
scientific               = ~r"-?(0|[1-9][0-9]*)(\.[0-9]+)?[eE]-?[0-9]+"
list                     = "[" _? (value (_ value)* _?)? "]"
object                   = "{" _? (pair (_ pair)* _?)? "}"

identifier               = ~"[_a-zA-Z][_a-zA-Z0-9]*"
type                     = (identifier (_? "!")?) / ("[" _? identifier (_? "!")? _? "]")
white                    = ~r"[\s,]+"
_                        = (white / comment)+
comment                  = "#" ~r".*(\n|$)"
'''
)


MISSING = object()


@dataclass
class ParsedField:
    """Represents a parsed field requested in an operation"""

    name: str
    alias: str = None  # type: ignore
    arguments: Optional[Dict[str, Any]] = None
    subfields: Optional[List["ParsedField"]] = None

    def __post_init__(self):
        if self.alias is None:
            self.alias = self.name


@dataclass
class ParsedEnum:
    """Represents an enum used as an argument value"""

    name: str


@dataclass
class ParsedVariable:
    """Represents a variable defined in or used by a query"""

    name: str


@dataclass
class ParsedOperation:
    """Represents a parsed operation from a NewQL query"""

    operation: str  # 'query' or 'mutation'
    fields: List[ParsedField]
    variables: Optional[List[str]] = None
    variable_defaults: Optional[Dict[str, Any]] = None
    name: Optional[str] = None


class NewQLVisitor(NodeVisitor):
    def visit_document(self, _, visited_children):
        output = {}
        for operation, _ in visited_children[1] or ():
            if operation.name in output:
                if operation.name is None:
                    message = "Multiple anonymous operations defined, only one allowed"
                else:
                    message = f"Operation name duplicated: '{operation.name}'"
                raise ValueError(message)
            output[operation.name] = operation

        return output

    def visit_operation(self, _, visited_children):
        operation = visited_children[0]
        # anonymous operation case
        if isinstance(operation, list):
            operation = ParsedOperation(operation="query", name=None, fields=operation)
        return operation

    def visit_explicit_operation(self, _, visited_children):
        operation, parsed_variables, _, fields = visited_children
        if parsed_variables:
            variables, variable_defaults = parsed_variables[0]
        else:
            variables = variable_defaults = None

        return ParsedOperation(operation, fields, variables, variable_defaults)

    def visit_named_explicit_operation(self, _, visited_children):
        operation, _, name, parsed_variables, _, fields = visited_children
        if parsed_variables:
            variables, variable_defaults = parsed_variables[0]
        else:
            variables = variable_defaults = None

        return ParsedOperation(operation, fields, variables, variable_defaults, name)

    def visit_operation_variables(self, _, visited_children):
        variable, default_value = visited_children[3]

        variables = [variable]
        variable_defaults = {}
        if default_value is not MISSING:
            variable_defaults[variable] = default_value

        for variable_match in visited_children[4] or ():
            variable, default_value = variable_match[1]
            if variable in variables:
                raise ValueError(f"Duplicate variable in operation: '{variable}'")
            variables.append(variable)
            if default_value is not MISSING:
                variable_defaults[variable] = default_value

        return variables, variable_defaults

    def visit_operation_variable(self, _, visited_children):
        variable_name = visited_children[1]
        # variable_type = visited_children[5]
        variable_default = visited_children[6]
        if variable_default is None:
            default_value = MISSING
        else:
            default_value = variable_default[0][3]

        return variable_name, default_value

    def visit_operation_type(self, node, _):
        return node.text

    def visit_fields(self, _, visited_children):
        fields = [visited_children[2]]

        additional_fields = visited_children[3]
        if additional_fields:
            for (_, field) in additional_fields:
                fields.append(field)

        return fields

    def visit_field(self, _, visited_children):
        name = visited_children[1]
        alias = visited_children[0]
        if alias is None:
            alias = name
        else:
            alias = alias[0]

        arguments = visited_children[2]
        if arguments is not None:
            arguments = arguments[0]

        subfields = visited_children[3]
        if subfields is not None:
            subfields = subfields[0][1]

        return ParsedField(name=name, alias=alias, arguments=arguments, subfields=subfields)

    def visit_field_alias(self, _, visited_children):
        return visited_children[0]

    def visit_field_arguments(self, _, visited_children):
        key, value = visited_children[3]
        arguments = {key: value}
        for additional_argument in visited_children[4] or ():
            key, value = additional_argument[1]
            if key in arguments:
                raise ValueError(f"Argument found twice in feature: '{key}'")
            arguments[key] = value

        return arguments

    def visit_pair(self, _, visited_children):
        argument_name = visited_children[0]
        value = visited_children[4]
        return argument_name, value

    def visit_value(self, _, visited_children):
        return visited_children[0]

    def visit_boolean(self, node, _):
        return node.text == "true"

    def visit_null(self, *_):
        return None

    def visit_string(self, node, _):
        return literal_eval(node.text)

    def visit_variable(self, _, visited_children):
        return ParsedVariable(visited_children[1])

    def visit_enum(self, _, visited_children):
        return ParsedEnum(visited_children[0])

    def visit_integer(self, node, _):
        return int(node.text)

    def visit_number(self, _, visited_children):
        return float(visited_children[0].text)

    def visit_list(self, _, visited_children):
        data = visited_children[2]
        if data is None:
            return []
        else:
            data = data[0]

        result = [data[0]]
        for _, additional in data[1] or ():
            result.append(additional)

        return result

    def visit_object(self, _, visited_children):
        data = visited_children[2]
        if data is None:
            return {}
        else:
            data = data[0]

        key, value = data[0]
        result = {key: value}
        for _, (key, value) in data[1] or ():
            if key in result:
                raise ValueError(f"Key found twice in object: '{key}'")
            result[key] = value

        return result

    def visit_identifier(self, node, _):
        return node.text

    def visit_type(self, _, visited_children):
        data = visited_children[0]
        if isinstance(data[0], str):
            return data[0]
        else:
            return data[2]

    def generic_visit(self, node, visited_children):
        if not visited_children and node.text == "":
            return None

        return visited_children or node


# TODO: replace VisitationError with the original errors (instead of stringifying them)
def parse_document(query_string: str) -> Dict[str, ParsedOperation]:
    visitor = NewQLVisitor()
    tree = grammar.parse(query_string)
    return visitor.visit(tree)


def parse_query(query_string: str, name: str = None) -> ParsedOperation:
    document = parse_document(query_string)
    if name is None:
        if len(document) > 1:
            raise ValueError("Multiple operations defined but none selected")

        return next(iter(document.values()))

    if name not in document:
        raise ValueError(f"Operation with name '{name}' not defined in document")

    return document[name]



from enum import Enum
from typing import List, Optional

from newql import ExecutionContext, Schema, field


class UserType(Enum):
    ADMIN = "admin"
    USER = "user"


class User:
    name = field(type=str, doc="The full name of the user")
    email = field(type=str, doc="The users contact email")
    age = field(type=int, doc="The current age of the user")

    @field
    def user_type(user: dict, _context) -> str:  # type: ignore
        return user["type"].value

    @field
    def phone(user: dict, context, area_code: bool = True) -> str:  # type: ignore
        """The user's contact phone number

        Args:
            area_code: Whether or not to include the area code
        """
        if area_code:
            return user["area_code"] + "-" + user["phone"]
        else:
            return user["phone"]

    @field
    def custom_arg(user, context, argument: str) -> Optional[str]:
        return argument


USERS = {
    "U1": {
        "name": "John Smith",
        "email": "johnsmith@example.com",
        "age": 34,
        "area_code": "555",
        "phone": "555-1234",
        "type": UserType.ADMIN,
    },
    "U2": {
        "name": "Jane Doe",
        "email": "janedoe@example.com",
        "age": 31,
        "area_code": "555",
        "phone": "555-9876",
        "type": UserType.USER,
    },
    "U-buggy-phone": {
        "name": "Adam Appleseed",
        "email": "adamappleseed@example.com",
        "age": 28,
        "area_code": None,
        "phone": "555-0000",
        "type": UserType.USER,
    },
}


class UserQuery:
    @field(type=User)
    def user(_, context, user_id: str) -> dict:
        if user_id in USERS:
            return USERS[user_id]
        else:
            raise ValueError(f"User '{user_id}' not found")

    @field(type=List[User])
    def users(_, context, user_ids: List[str] = None, user_type: UserType = None) -> List[dict]:
        return [
            user
            for user_id, user in USERS.items()
            if (user_ids is None or user_id in user_ids) and (not user_type or user["type"] == user_type)
        ]


user_schema = Schema(UserQuery)


class Product:
    """Represents a product in stock"""

    product_id = field(type=int)
    product_name = field(type=str)
    price = field(type=float)


class ProductQuery:
    # pass 'type' here since the function returns a dictionary,
    # but we want to actually resolve to a Product
    @field(type=Product)
    def product(_, context: ExecutionContext, product_id: int) -> dict:
        # The "Args" section of the docstring is parsed out to add
        # docs to the arguments (can be seen in GraphiQL)
        """Find a specific product by id

        Args:
            product_id: The ID of the product to find
        """

        products = {1: ("Product1", 49.99), 2: ("Product2", 94.49)}
        if product_id in products:
            name, price = products[product_id]
            return {"product_id": product_id, "product_name": name, "price": price}
        else:
            raise ValueError(f"Product not found: {product_id}")


product_schema = Schema(ProductQuery)



from importlib import resources

from .context import ExecutionContext
from .errors import NewQLError, QueryError
from .field import Field, field
from .parse import ParsedEnum, ParsedField, ParsedOperation, ParsedVariable, parse_document, parse_query
from .schema import Schema

VERSION = resources.read_text("newql", "VERSION").strip()

__all__ = [
    "ExecutionContext",
    "Field",
    "NewQLError",
    "ParsedOperation",
    "ParsedEnum",
    "ParsedField",
    "ParsedVariable",
    "QueryError",
    "Schema",
    "VERSION",
    "field",
    "parse_document",
    "parse_query",
]



import inspect
import re
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional, Union, get_type_hints

from varname import varname  # type: ignore

from .context import ExecutionContext

FIELD_NAME_PATTERN = re.compile("[_a-zA-Z][_a-zA-Z0-9]*")


@dataclass
class Field:
    name: str
    resolver: Callable
    field_type: type
    args: Dict[str, type]
    doc: Optional[str] = None

    # don't lose the ability to call functions wrapped with @field
    def __call__(self, *args, **kwargs):
        return self.resolver(*args, **kwargs)


def default_resolver(field_name: str) -> Callable[[Any, ExecutionContext], Any]:
    def dict_or_attribute_resolver(obj, _):
        if isinstance(obj, Mapping):
            return obj[field_name]
        else:
            return getattr(obj, field_name)

    return dict_or_attribute_resolver


def _extract_field_args(resolver: Callable, explicit_args: Optional[Dict[str, type]]) -> Dict[str, type]:
    E = f"Resolver '{resolver.__name__}' "

    arg_spec = inspect.getfullargspec(resolver)
    if arg_spec.varargs is None and len(arg_spec.args) < 2:
        raise ValueError(E + "must accept at least two positional arguments")

    parameters = set(arg_spec.args[2:] + arg_spec.kwonlyargs)
    field_args = {arg: arg_type for arg, arg_type in arg_spec.annotations.items() if arg in parameters}

    if explicit_args is None:
        all_args = field_args
    else:
        fixed_explicit_args = {arg: arg_type for arg, arg_type in explicit_args.items() if arg_type}
        all_args = field_args.copy()
        all_args.update(fixed_explicit_args)

    missing_hints = parameters - all_args.keys()
    if missing_hints:
        raise ValueError(E + f"missing type hints for {sorted(missing_hints)}")

    if arg_spec.varkw and explicit_args is None:
        raise ValueError(
            f"Must explicitly set 'args' for resolver '{resolver.__name__}' with argument '**{arg_spec.varkw}'"
        )

    # if present, make sure no parameters of the function
    # are missing from the explicit arguments
    if explicit_args is not None:
        extra_explicit = explicit_args.keys() - parameters
        if arg_spec.varkw is None and extra_explicit:
            raise ValueError(E + f"does not have all arguments in 'args': {sorted(extra_explicit)}")

    return all_args


def field(
    _resolver_or_name: Union[str, Callable] = None,
    *,
    name: str = None,
    resolver: Callable = None,
    type: type = None,
    args: Dict[str, type] = None,
    doc: str = None,
) -> Union[Field, Callable[..., Field]]:
    '''
    A decorator to turn a resolver function into a field. Can also be used
    to construct a Field directly, defaulting to a resolver that tries to find
    the field as an attribute (or key in a dictionary)

    When called as a constructor, the first argument must be the field name (not passed
    in as `name="..."`, as that is reserved for higher order decorators:

        class MyType:
            # uses the default resolver
            my_field = field(type=str)

            # uses custom resolver
            my_custom_field = field(type=int, resolver=my_resolver)

            # can pass explicit name
            anything = field("something", type=str)

    Can also be called as a higher order decorator to override field information:

        @field(name="actual_name", type=str, doc="Documentation about the string")
        def _gibberish(parent, context: ExecutionContext):
            return "foo"

        # equivalent to the following
        @field
        def actual_name(parent, context) -> str:
            """Documentation about the field"""
            return "foo"

    These higher order decorators are particularly useful for fields
    that return nested fields, for example:

        class Nested:
            @field
            def first_data_word(parent, _) -> str:
                return parent["some_data"].split()[0]

        class Parent:
            # type=Nested is necessary (at least when using mypy) since the function
            # returns a dict, but the field type is actually Nested
            @field(type=Nested)
            def nested(_, _) -> dict:
                return {"some_data": "goes here"}
    '''

    if _resolver_or_name is None:
        # either called as a higher order decorator `@field(...)`,
        # or as a constructor with implicit name `my_field = field(type=str)`
        if name is None:
            # if called as a field, we will be able to extract varname
            try:
                name = name or varname()  # type: ignore
            except Exception:
                # otherwise, varname() will raise an exception and name will still be None
                pass
            else:
                # only set resolver in the case where we successfully extracted varname
                # (and so field was called as a constructor, missing the name)
                resolver = resolver or default_resolver(name)
    elif isinstance(_resolver_or_name, str):
        # calling `field` as a constructor, use provided resolver or the default one
        if name is not None:
            raise ValueError("Cannot specify field name with both positional and keyword argument")
        name = _resolver_or_name
        resolver = resolver or default_resolver(name)
    else:
        # calling `field` as a flat decorator
        if resolver is not None:
            raise ValueError("Cannot specify resolver with both positional and keyword argument")
        resolver = _resolver_or_name

    def decorator(resolver: Callable) -> Field:
        field_name = name or resolver.__name__
        if not FIELD_NAME_PATTERN.fullmatch(field_name):
            raise ValueError(f"Field name must match {FIELD_NAME_PATTERN.pattern}, found '{field_name}'")

        field_type = type or get_type_hints(resolver).get("return")
        if field_type is None:
            raise ValueError(
                f"No return value type hint on resolver and `type` parameter not present for '{field_name}'"
            )

        field_args = _extract_field_args(resolver, args)
        field_doc = doc or resolver.__doc__

        return Field(field_name, resolver, field_type, field_args, field_doc)

    return decorator if resolver is None else decorator(resolver)



import argparse
import json
from http.server import BaseHTTPRequestHandler, HTTPServer
from importlib import import_module

from newql import Schema

GRAPHIQL = """
<html>
    <head>
        <meta charset="UTF-8">
        <title>NewQL Dev Server (GraphiQL)</title>
        <link href="https://unpkg.com/graphiql/graphiql.min.css" rel="stylesheet" />
    </head>
    <body style="margin: 0;">
        <div id="graphiql" style="height: 100vh;"></div>
        <script crossorigin src="https://unpkg.com/react/umd/react.production.min.js"></script>
        <script crossorigin src="https://unpkg.com/react-dom/umd/react-dom.production.min.js"></script>
        <script crossorigin src="https://unpkg.com/graphiql/graphiql.min.js"></script>
        <script>
         const fetcher = graphQLParams =>
             fetch('/', {
                 method: 'post',
                 headers: {'Content-Type': 'application/json'},
                 body: JSON.stringify(graphQLParams),
             }).then(response => response.json()).catch(() => response.text());

         ReactDOM.render(React.createElement(GraphiQL, {fetcher}), document.getElementById('graphiql'));
        </script>
    </body>
</html>
"""


def run(schema: Schema, address: str, port: int):
    class Handler(BaseHTTPRequestHandler):
        def _set_headers(self):
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()

        def do_GET(self):
            self._set_headers()
            self.wfile.write(GRAPHIQL.encode("utf-8"))

        def do_POST(self):
            self._set_headers()
            request = json.loads(self.rfile.read(int(self.headers["Content-Length"])))
            result = schema.execute(request["query"], request.get("variable"), request.get("operation"))
            self.wfile.write(json.dumps(result).encode("utf-8"))

    print(f"Starting NewQL Dev Server at {address}:{port}")
    HTTPServer((address, port), Handler).serve_forever()


def _import_schema(import_path: str) -> Schema:
    module_path, schema = import_path.rsplit(".", 1)
    module = import_module(module_path)
    return getattr(module, schema)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run a simple HTTP server")
    parser.add_argument(
        "-l",
        "--listen",
        default="localhost",
        help="Specify the IP address on which the server listens",
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=8000,
        help="Specify the port on which the server listens",
    )
    parser.add_argument(
        "-s",
        "--schema",
        type=str,
        default="newql.example.user_schema",
        help="The full import name of the schema to server",
    )
    args = parser.parse_args()
    schema = _import_schema(args.schema)
    run(schema, args.listen, args.port)



from functools import partial
from typing import Any, Dict, Optional, TypeVar

from .context import ExecutionContext
from .errors import QueryError
from .field import Field
from .introspection import check_list, check_optional, get_list_type, introspect_type
from .parse import ParsedEnum, ParsedField, ParsedOperation, ParsedVariable, parse_query

SKIP_FIELD = object()


def _validate_variables(variables: Optional[Dict[str, Any]], operation: ParsedOperation) -> Dict[str, Any]:
    all_variables = operation.variable_defaults or {}
    all_variables.update(variables or {})

    operation_variables = set(operation.variables or ())

    missing_variables = operation_variables - all_variables.keys()
    if missing_variables:
        raise QueryError(f"Required variables not provided: {sorted(missing_variables)}")

    return all_variables


def _resolve_result(context: ExecutionContext, field: type, requested_field: ParsedField, parent: Any):
    result = {}
    for subfield in requested_field.subfields or ():
        sub_value = _resolve_field(context, field, subfield, parent)
        if sub_value is not SKIP_FIELD:
            result[subfield.alias] = sub_value
    return result


def _resolve_arguments(context: ExecutionContext, data, arg_type):
    # TODO: make sure this holds up for deeply nested dict/list/optional argument types

    arg_type, _ = check_optional(arg_type)

    if isinstance(data, dict):
        return {key: _resolve_arguments(context, value, arg_type[key]) for key, value in data.items()}
    elif isinstance(data, list):
        list_type = get_list_type(arg_type)
        return [_resolve_arguments(context, item, list_type) for item in data]
    elif isinstance(data, ParsedVariable):
        return context.variables[data.name]
    elif isinstance(data, ParsedEnum):
        return getattr(arg_type, data.name, data.name)
    else:
        return data


def _resolve_field(
    context: ExecutionContext,
    fields_class: type,
    requested_field: ParsedField,
    parent: Any = None,
):
    context.push(requested_field)
    field = getattr(fields_class, requested_field.name, None)
    if not isinstance(field, Field):
        context.error("Unknown field requested", requested_field)
        return SKIP_FIELD

    raw_args = requested_field.arguments or {}
    arguments: dict = _resolve_arguments(context, raw_args, field.args)  # type: ignore

    try:
        resolved = field.resolver(parent, context, **arguments)
    except Exception as e:
        context.error(e, requested_field)
        return None

    if resolved is None or not requested_field.subfields:
        context.pop(requested_field)
        return resolved

    field_type = field.field_type
    if check_list(field_type):
        list_type = get_list_type(field_type)
        if not list_type or isinstance(list_type, TypeVar):
            result = resolved
        else:
            item_resolver = partial(_resolve_result, context, list_type, requested_field)
            result = list(map(item_resolver, resolved))
    else:
        result = _resolve_result(context, field_type, requested_field, resolved)

    context.pop(requested_field)
    return result


class Schema:
    def __init__(self, query_class: type = None, mutation_class: type = None):
        self.query = query_class
        self.mutation = mutation_class

    def execute(self, query: str, variables: Optional[Dict[str, Any]] = None, operation_name: str = None):
        try:
            operation = parse_query(query, operation_name)
        except Exception as e:
            # The default introspection query in GraphiQL has fragments
            # which NewQL doesn't support, so fallback to check for __schema
            # and just return the introspection if suspected
            if "__schema" in query:
                operation = parse_query("{ __schema }")
            else:
                return {"data": None, "errors": str(e)}

        operation_class = getattr(self, operation.operation)
        if operation_class is None:
            raise QueryError(f"Operation '{operation.operation}' is not defined for this schema")

        fixed_variables = _validate_variables(variables, operation)

        context = ExecutionContext(operation, fixed_variables, None)  # type: ignore
        resolved = {}
        for requested_field in operation.fields:
            # if __schema is present, just return the entire introspection
            # (mostly just used for the GraphiQL intrgration)
            if requested_field.name == "__schema":
                value = self.introspect()
            else:
                value = _resolve_field(context, operation_class, requested_field)
            if value is not SKIP_FIELD:
                resolved[requested_field.alias] = value

        result: dict = {"data": resolved}
        if context.errors:
            result["errors"] = context.errors
        return result

    def _introspect(self, t: Optional[type], types_by_class: dict, types_by_name: dict) -> Optional[dict]:
        if t is None:
            return None

        introspected = introspect_type(t, types_by_class, types_by_name, is_operation=True)
        if not introspected:
            return None

        return {"name": introspected["name"]}

    def introspect(self):
        types_by_class = {}
        types_by_name = {}

        return {
            "queryType": self._introspect(self.query, types_by_class, types_by_name),
            "mutationType": self._introspect(self.mutation, types_by_class, types_by_name),
            "types": list(types_by_name.values()),
        }



from pathlib import Path

from setuptools import setup


ROOT_DIRECTORY = Path(__file__).resolve().parent

description = "A simplified GraphQL-esque library"
readme = (ROOT_DIRECTORY / "README.md").read_text()
changelog = (ROOT_DIRECTORY / "CHANGELOG.md").read_text()
long_description = readme + "\n\n" + changelog

version = (ROOT_DIRECTORY / "newql" / "VERSION").read_text().strip()


DEV_REQUIRES = [
    "black==21.6b0",
    "coverage==5.5",
    "flake8==3.9.2",
    "flake8-bugbear==21.4.3",
    "isort==5.9.1",
    "mypy==0.910",
    "pytest==6.2.4",
    "pytest-cov==2.12.1",
    "twine==3.4.1",
]

setup(
    name="newql",
    version=version,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    author="Naya Verdier",
    url="https://github.com/nayaverdier/newql",
    license="MIT",
    packages=["newql"],
    install_requires=[
        "parsimonious>=0.8",
        "typing_inspect;python_version<='3.7'",
        "varname~=0.6",
    ],
    python_requires=">=3.7",
    extras_require={
        "dev": DEV_REQUIRES,
    },
    include_package_data=True,
)



from dataclasses import dataclass
from typing import Any, Dict, List, Union

from .parse import ParsedField, ParsedOperation


@dataclass
class ExecutionContext:
    operation: ParsedOperation
    variables: Dict[str, Any]
    current_field: ParsedField

    def __post_init__(self):
        self.errors: Dict[str, Any] = {}
        self.path: List[str] = []

    def push(self, field: ParsedField):
        self.current_field = field
        self.path.append(field.alias)

    def pop(self, field: ParsedField):
        assert field.alias == self.path.pop()

    def error(self, error: Union[str, Exception], field: ParsedField):
        location = self.errors
        for p in self.path[:-1]:
            location = location.setdefault(p, {})
        location[self.path[-1]] = str(error)

        self.pop(field)
